from copy import deepcopy
from typing import Iterable

import lxml.etree as ET

from burst2safe.base import Annotation
from burst2safe.utils import BurstInfo


class Calibration(Annotation):
    """Class representing a calibration XML."""

    def __init__(self, burst_infos: Iterable[BurstInfo], image_number: int):
        """Create a calibration object.

        Args:
            burst_infos: List of BurstInfo objects.
            image_number: Image number.
        """
        super().__init__(burst_infos, 'calibration', image_number)
        self.calibration_information = None
        self.calibrattion_vector_list = None

    def create_calibration_information(self):
        """Create the calibration information."""
        calibration_information = [calibration.find('calibrationInformation') for calibration in self.inputs][0]
        self.calibration_information = deepcopy(calibration_information)

    def create_calibration_vector_list(self):
        """Create the calibration vector list."""
        self.calibration_vector_list = self.merge_lists('calibrationVectorList')

    def assemble(self):
        """Assemble the calibration object components."""
        self.create_ads_header()
        self.create_calibration_information()
        self.create_calibration_vector_list()

        calibration = ET.Element('calibration')
        calibration.append(self.ads_header)
        calibration.append(self.calibration_information)
        calibration.append(self.calibration_vector_list)
        calibration_tree = ET.ElementTree(calibration)

        ET.indent(calibration_tree, space='  ')
        self.xml = calibration_tree



from copy import deepcopy
from pathlib import Path
from typing import List

import lxml.etree as ET
import numpy as np
from shapely.geometry import Polygon

from burst2safe.utils import calculate_crc16


class Manifest:
    """Class representing a SAFE manifest."""

    def __init__(
        self,
        content_units: List[ET.Element],
        metadata_objects: List[ET.Element],
        data_objects: List[ET.Element],
        bbox: Polygon,
        template_manifest: ET.Element,
    ):
        """Initialize a Manifest object.

        Args:
            content_units: A list of contentUnit elements
            metadata_objects: A list of metadataObject elements
            data_objects: A list of dataObject elements
            bbox: The bounding box of the product
            template_manifest: The template manifest to generate unalterd metadata from
        """
        self.content_units = content_units
        self.metadata_objects = metadata_objects
        self.data_objects = data_objects
        self.bbox = bbox
        self.template = template_manifest

        safe_ns = 'http://www.esa.int/safe/sentinel-1.0'
        self.namespaces = {
            'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'gml': 'http://www.opengis.net/gml',
            'xfdu': 'urn:ccsds:schema:xfdu:1',
            'safe': safe_ns,
            's1': f'{safe_ns}/sentinel-1',
            's1sar': f'{safe_ns}/sentinel-1/sar',
            's1sarl1': f'{safe_ns}/sentinel-1/sar/level-1',
            's1sarl2': f'{safe_ns}/sentinel-1/sar/level-2',
            'gx': 'http://www.google.com/kml/ext/2.2',
        }
        self.version = 'esa/safe/sentinel-1.0/sentinel-1/sar/level-1/slc/standard/iwdp'

        # Updated by methods
        self.information_package_map = None
        self.metadata_section = None
        self.data_object_section = None
        self.xml = None
        self.path = None
        self.crc = None

    def create_information_package_map(self):
        """Create the information package map."""
        xdfu_ns = self.namespaces['xfdu']
        information_package_map = ET.Element(f'{{{xdfu_ns}}}informationPackageMap')
        parent_content_unit = ET.Element(
            f'{{{xdfu_ns}}}contentUnit',
            unitType='SAFE Archive Information Package',
            textInfo='Sentinel-1 IW Level-1 SLC Product',
            dmdID='acquisitionPeriod platform generalProductInformation measurementOrbitReference measurementFrameSet',
            pdiID='processing',
        )
        for content_unit in self.content_units:
            parent_content_unit.append(content_unit)

        information_package_map.append(parent_content_unit)
        self.information_package_map = information_package_map

    def create_metadata_section(self):
        """Create the metadata section."""
        metadata_section = ET.Element('metadataSection')
        for metadata_object in self.metadata_objects:
            metadata_section.append(metadata_object)

        ids_to_keep = [
            'processing',
            'platform',
            'measurementOrbitReference',
            'generalProductInformation',
            'acquisitionPeriod',
            'measurementFrameSet',
        ]
        section = 'metadataSection'
        [metadata_section.append(deepcopy(x)) for x in self.template.find(section) if x.get('ID') in ids_to_keep]

        new_coords = [(np.round(y, 6), np.round(x, 6)) for x, y in self.bbox.exterior.coords]
        # TODO: only works for descending
        new_coords = [new_coords[2], new_coords[3], new_coords[0], new_coords[1]]
        new_coords = ' '.join([f'{x},{y}' for x, y in new_coords])
        coordinates = metadata_section.find('.//{*}coordinates')
        coordinates.text = new_coords

        self.metadata_section = metadata_section

    def create_data_object_section(self):
        """Create the data object section."""
        self.data_object_section = ET.Element('dataObjectSection')
        for data_object in self.data_objects:
            self.data_object_section.append(data_object)

    def assemble(self):
        """Assemble the components of the SAFE manifest."""
        self.create_information_package_map()
        self.create_metadata_section()
        self.create_data_object_section()

        manifest = ET.Element('{%s}XFDU' % self.namespaces['xfdu'], nsmap=self.namespaces)
        manifest.set('version', self.version)
        manifest.append(self.information_package_map)
        manifest.append(self.metadata_section)
        manifest.append(self.data_object_section)
        manifest_tree = ET.ElementTree(manifest)

        ET.indent(manifest_tree, space='  ')
        self.xml = manifest_tree

    def write(self, out_path: Path, update_info: bool = True) -> None:
        """Write the SAFE manifest to a file. Optionally update the path and CRC.

        Args:
            out_path: The path to write the manifest to
            update_info: Whether to update the path and CRC
        """
        self.xml.write(out_path, pretty_print=True, xml_declaration=True, encoding='utf-8')
        if update_info:
            self.path = out_path
            self.crc = calculate_crc16(self.path)



from datetime import datetime
from pathlib import Path

import asf_search
import pytest

from burst2safe.utils import BurstInfo


TEST_DIR = Path(__file__).parent


def pytest_configure(config):
    config.addinivalue_line('filterwarnings', 'ignore::RuntimeWarning')


@pytest.fixture
def test_data_dir():
    return TEST_DIR / 'test_data'


@pytest.fixture
def test_data_xml(test_data_dir):
    return test_data_dir / 'S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85_VV.xml'


@pytest.fixture
def xsd_dir():
    xsd_dir = Path(__file__).parent.parent / 'src' / 'burst2safe' / 'data'
    return xsd_dir


@pytest.fixture
def search_result1():
    product = asf_search.ASFProduct()
    product.umm = {'InputGranules': ['S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85']}
    product.properties.update(
        {
            'fileID': 'S1_136231_IW2_20200604T022312_VV_7C85-BURST',
            'flightDirection': 'ascending',
            'polarization': 'vv',
            'orbit': 123,
            'url': 'https://example.com/foo.zip',
            'additionalUrls': ['https://example.com/foo.xml'],
            'burst': {
                'subswath': 'IW2',
                'relativeBurstID': 123456,
                'burstIndex': 7,
            },
        }
    )
    # If an ASFSearchResults object is needed, uncomment the following lines:
    # results = asf_search.ASFSearchResults([product])
    # results.searchComplete = True
    return product


@pytest.fixture
def burst_info1(test_data_xml):
    burst_info = BurstInfo(
        granule='S1_136231_IW2_20200604T022312_VV_7C85-BURST',
        slc_granule='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
        swath='IW2',
        polarization='VV',
        burst_id=136231,
        burst_index=7,
        direction='DESCENDING',
        absolute_orbit=32861,
        date=datetime(2020, 6, 4, 2, 23, 12),
        data_url='https://sentinel1-burst.asf.alaska.edu/S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85/IW2/VV/7.tiff',
        data_path=Path(''),
        metadata_url='https://sentinel1-burst.asf.alaska.edu/S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85/IW2/VV/7.xml',
        metadata_path=test_data_xml,
        start_utc=datetime(2020, 6, 4, 2, 23, 12, 933265),
        stop_utc=datetime(2020, 6, 4, 2, 23, 16, 37825),
        length=1510,
    )
    return burst_info


@pytest.fixture
def burst_info2(test_data_xml):
    burst_info = BurstInfo(
        granule='S1_136232_IW2_20200604T022315_VV_7C85-BURST',
        slc_granule='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
        swath='IW2',
        polarization='VV',
        burst_id=136232,
        burst_index=8,
        direction='DESCENDING',
        absolute_orbit=32861,
        date=datetime(2020, 6, 4, 2, 23, 15),
        data_url='https://sentinel1-burst.asf.alaska.edu/S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85/IW2/VV/8.tiff',
        data_path=Path(''),
        metadata_url='https://sentinel1-burst.asf.alaska.edu/S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85/IW2/VV/8.xml',
        metadata_path=test_data_xml,
        start_utc=datetime(2020, 6, 4, 2, 23, 15, 697989),
        stop_utc=datetime(2020, 6, 4, 2, 23, 18, 802549),
        length=1510,
    )
    return burst_info


@pytest.fixture
def burst_infos(burst_info1, burst_info2):
    return [burst_info1, burst_info2]



from datetime import datetime
from pathlib import Path
from typing import Iterable

from shapely.geometry import MultiPoint, Polygon

from burst2safe.calibration import Calibration
from burst2safe.measurement import Measurement
from burst2safe.noise import Noise
from burst2safe.product import Product
from burst2safe.utils import BurstInfo


class Swath:
    """Class representing a single swath (and polarization) of a SAFE file."""

    def __init__(self, burst_infos: Iterable[BurstInfo], safe_path: Path, image_number: int):
        """Initialize a Swath object."""
        self.check_burst_group_validity(burst_infos)
        self.burst_infos = sorted(burst_infos, key=lambda x: x.burst_id)
        self.safe_path = safe_path
        self.image_number = image_number
        self.swath = self.burst_infos[0].swath
        self.polarization = self.burst_infos[0].polarization

        self.name = self.get_name()

        self.measurement_name = self.safe_path / 'measurement' / f'{self.name}.tiff'
        self.product_name = self.safe_path / 'annotation' / f'{self.name}.xml'
        self.noise_name = self.safe_path / 'annotation' / 'calibration' / f'noise-{self.name}.xml'
        self.calibration_name = self.safe_path / 'annotation' / 'calibration' / f'calibration-{self.name}.xml'

        # Set on write
        self.bbox = None

    @staticmethod
    def check_burst_group_validity(burst_infos: Iterable[BurstInfo]):
        """Check that the burst group is valid.

        The burst group must:
        - Not contain duplicate granules
        - Have the same absolute orbit
        - Be from the same swath
        - Have the same polarization
        - Have consecutive burst IDs

        Args:
            burst_infos: A list of BurstInfo objects
        """
        granules = [x.granule for x in burst_infos]
        duplicates = list(set([x for x in granules if granules.count(x) > 1]))
        if duplicates:
            raise ValueError(f'Found duplicate granules: {duplicates}.')

        orbits = set([x.absolute_orbit for x in burst_infos])
        if len(orbits) != 1:
            raise ValueError(f'All bursts must have the same absolute orbit. Found: {orbits}.')

        swaths = set([x.swath for x in burst_infos])
        if len(swaths) != 1:
            raise ValueError(f'All bursts must be from the same swath. Found: {swaths}.')

        polarizations = set([x.polarization for x in burst_infos])
        if len(polarizations) != 1:
            raise ValueError(f'All bursts must have the same polarization. Found: {polarizations}.')

        burst_ids = [x.burst_id for x in burst_infos]
        burst_ids.sort()
        if burst_ids != list(range(min(burst_ids), max(burst_ids) + 1)):
            raise ValueError(f'All bursts must have consecutive burst IDs. Found: {burst_ids}.')

    def get_name(self) -> str:
        """Get the name of the swath. Will be used to name constituent output files.

        Returns:
            The name of the swath
        """
        swath = self.swath.lower()
        pol = self.polarization.lower()
        start = datetime.strftime(min([x.start_utc for x in self.burst_infos]), '%Y%m%dt%H%M%S')
        stop = datetime.strftime(max([x.stop_utc for x in self.burst_infos]), '%Y%m%dt%H%M%S')

        safe_name = self.safe_path.name
        platfrom, _, _, _, _, _, _, orbit, data_take, _ = safe_name.lower().split('_')
        swath_name = f'{platfrom}-{swath}-slc-{pol}-{start}-{stop}-{orbit}-{data_take}-{self.image_number:03d}'
        return swath_name

    def get_bbox(self):
        """Get the bounding box of the swath using the Product GCPs.

        Returns:
            The bounding box of the swath
        """
        points = MultiPoint([(gcp.x, gcp.y) for gcp in self.product.gcps])
        min_rotated_rect = points.minimum_rotated_rectangle
        bbox = Polygon(min_rotated_rect.exterior)
        return bbox

    def assemble(self):
        """Assemble the components of the Swath."""
        self.product = Product(self.burst_infos, self.image_number)
        self.noise = Noise(self.burst_infos, self.image_number)
        self.calibration = Calibration(self.burst_infos, self.image_number)
        self.annotations = [self.product, self.noise, self.calibration]
        for component in self.annotations:
            component.assemble()

        self.measurement = Measurement(self.burst_infos, self.product.gcps, self.image_number)

    def write(self, update_info: bool = True):
        """Write the Swath componets to the SAFE directory.

        Args:
            update_info: Whether to update the bounding box of the Swath
        """
        self.measurement.write(self.measurement_name)
        self.product.update_data_stats(self.measurement.data_mean, self.measurement.data_std)
        self.product.write(self.product_name)
        self.noise.write(self.noise_name)
        self.calibration.write(self.calibration_name)
        if update_info:
            self.bbox = self.get_bbox()

    def create_manifest_components(self):
        """Create the manifest components for the Swath."""
        self.manifest_components = {
            'content_unit': [],
            'metadata_object': [],
            'data_object': [],
            'content_unit_measurement': [],
            'data_object_measurement': [],
        }
        for annotation in self.annotations:
            content, metadata, data = annotation.create_manifest_components()
            self.manifest_components['content_unit'].append(content)
            self.manifest_components['metadata_object'].append(metadata)
            self.manifest_components['data_object'].append(data)

        content, data = self.measurement.create_manifest_components()
        self.manifest_components['content_unit_measurement'].append(content)
        self.manifest_components['data_object_measurement'].append(data)



from collections import namedtuple

import pytest
from shapely.geometry import Polygon

from burst2safe.safe import Safe


class TestSafe:
    def test_init(self, burst_infos, tmp_path):
        safe = Safe(burst_infos, work_dir=tmp_path)
        assert safe.burst_infos == burst_infos
        assert safe.work_dir == tmp_path
        assert safe.swaths == []

    def test_check_group_validity(self):
        BurstStub = namedtuple('BurstStub', ['granule', 'absolute_orbit', 'swath', 'polarization', 'burst_id'])
        burst1 = BurstStub(granule='S1A_20210101_1_VV', absolute_orbit=1, swath='IW1', polarization='VV', burst_id=1)
        burst3 = BurstStub(granule='S1A_20210101_1_VH', absolute_orbit=1, swath='IW1', polarization='VH', burst_id=1)
        burst5 = BurstStub(granule='S1A_20210101_2_VV', absolute_orbit=1, swath='IW1', polarization='VV', burst_id=2)
        burst6 = BurstStub(granule='S1A_20210101_2_VH', absolute_orbit=1, swath='IW1', polarization='VH', burst_id=2)

        burst2 = BurstStub(granule='S1A_20210101_1_VV', absolute_orbit=1, swath='IW2', polarization='VV', burst_id=1)
        burst4 = BurstStub(granule='S1A_20210101_1_VH', absolute_orbit=1, swath='IW2', polarization='VH', burst_id=1)

        Safe.check_group_validity([burst1, burst2, burst3, burst4])

        burst_start = BurstStub(
            granule='S1A_IW_SLC_20210101_0_VH', absolute_orbit=1, swath='IW1', polarization='VH', burst_id=0
        )
        different_polarization_starts = [burst_start, burst1, burst3, burst5, burst6]
        with pytest.raises(ValueError, match='Polarization groups in swath .* start burst id.*'):
            Safe.check_group_validity(different_polarization_starts)

        burst_end = BurstStub(
            granule='S1A_IW_SLC_20210101_3_VH', absolute_orbit=1, swath='IW1', polarization='VH', burst_id=3
        )
        different_polarization_ends = [burst1, burst3, burst5, burst6, burst_end]
        with pytest.raises(ValueError, match='Polarization groups in swath .* end burst id.*'):
            Safe.check_group_validity(different_polarization_ends)

        swath_nonoverlap = [burst1, BurstStub(*burst2._replace(burst_id=3))]
        with pytest.raises(ValueError, match='Products from swaths IW1 and IW2 do not overlap'):
            Safe.check_group_validity(swath_nonoverlap)

    def test_get_name(self, burst_infos, tmp_path):
        safe = Safe(burst_infos, work_dir=tmp_path)
        golden_name = 'S1A_IW_SLC__1SSV_20200604T022312_20200604T022315_032861_03CE65_0000.SAFE'
        assert safe.get_name(safe.burst_infos) == golden_name

    def test_group_burst_infos(self):
        BurstStub = namedtuple('BurstStub', ['swath', 'polarization', 'burst_id'])
        burst1 = BurstStub(swath='IW1', polarization='VV', burst_id=1)
        burst2 = BurstStub(swath='IW1', polarization='VH', burst_id=1)
        burst3 = BurstStub(swath='IW1', polarization='VV', burst_id=2)
        burst4 = BurstStub(swath='IW1', polarization='VH', burst_id=2)
        burst5 = BurstStub(swath='IW2', polarization='VV', burst_id=1)
        burst6 = BurstStub(swath='IW2', polarization='VH', burst_id=1)
        burst7 = BurstStub(swath='IW2', polarization='VV', burst_id=2)
        burst8 = BurstStub(swath='IW2', polarization='VH', burst_id=2)

        burst_info_stubs = [burst1, burst2, burst3, burst4, burst5, burst6, burst7, burst8]
        grouped = Safe.group_burst_infos(burst_info_stubs)
        assert grouped['IW1']['VV'] == [burst1, burst3]
        assert grouped['IW1']['VH'] == [burst2, burst4]
        assert grouped['IW2']['VV'] == [burst5, burst7]
        assert grouped['IW2']['VH'] == [burst6, burst8]

    def test_get_bbox(self, burst_infos, tmp_path):
        safe = Safe(burst_infos, work_dir=tmp_path)

        polygon1 = Polygon([(1, 1), (1, 2), (2, 2), (2, 1), (1, 1)])
        polygon2 = Polygon([(2, 2), (2, 3), (3, 3), (3, 2), (2, 2)])
        polygon_merged = Polygon([(1, 1), (1, 3), (3, 3), (3, 1), (1, 1)])

        SwathStub = namedtuple('SwathStub', ['bbox'])
        safe.swaths = [SwathStub(polygon1), SwathStub(polygon2)]
        bbox = safe.get_bbox()
        assert bbox == polygon_merged

    def test_create_dir_structure(self, burst_infos, tmp_path):
        safe = Safe(burst_infos, work_dir=tmp_path)
        safe.create_dir_structure()
        assert safe.safe_path == tmp_path / safe.get_name(safe.burst_infos)
        assert (safe.safe_path / 'measurement').exists()
        assert (safe.safe_path / 'annotation').exists()
        assert (safe.safe_path / 'annotation' / 'calibration').exists()
        assert (safe.safe_path / 'support').exists()
        assert len(list((safe.safe_path / 'support').glob('*'))) == 5



from collections import namedtuple
from pathlib import Path

import pytest
from shapely.geometry import Polygon

from burst2safe.product import GeoPoint
from burst2safe.swath import Swath


class TestSwath:
    def test_init(self, burst_infos):
        safe_path = Path('./S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85.SAFE')
        swath = Swath(burst_infos[::-1], safe_path, 1)
        assert swath.burst_infos == burst_infos

        name = 's1a-iw2-slc-vv-20200604t022312-20200604t022318-032861-03ce65-001'

        assert swath.name == name
        assert swath.measurement_name.name == f'{name}.tiff'
        assert swath.measurement_name.parent.name == 'measurement'
        assert swath.product_name.name == f'{name}.xml'
        assert swath.product_name.parent.name == 'annotation'
        assert swath.noise_name.name == f'noise-{name}.xml'
        assert swath.noise_name.parent.name == 'calibration'
        assert swath.calibration_name.name == f'calibration-{name}.xml'
        assert swath.calibration_name.parent.name == 'calibration'

    def test_check_burst_group_validity(self):
        BurstStub = namedtuple('BurstStub', ['granule', 'absolute_orbit', 'swath', 'polarization', 'burst_id'])
        burst1 = BurstStub(granule='S1A_IW_SLC_20210101', absolute_orbit=1, swath='IW1', polarization='VV', burst_id=1)
        burst2 = BurstStub(granule='S1A_IW_SLC_20210102', absolute_orbit=1, swath='IW1', polarization='VV', burst_id=2)

        Swath.check_burst_group_validity([burst1, burst2])

        duplicates = [burst1, BurstStub(*burst2._replace(granule=burst1.granule))]
        with pytest.raises(ValueError, match='Found duplicate granules:.*'):
            Swath.check_burst_group_validity(duplicates)

        different_orbit = [burst1, BurstStub(*burst2._replace(absolute_orbit=2))]
        with pytest.raises(ValueError, match='All bursts must have the same absolute orbit. Found:.*'):
            Swath.check_burst_group_validity(different_orbit)

        different_swath = [burst1, BurstStub(*burst2._replace(swath='IW2'))]
        with pytest.raises(ValueError, match='All bursts must be from the same swath. Found:.*'):
            Swath.check_burst_group_validity(different_swath)

        different_polarization = [burst1, BurstStub(*burst2._replace(polarization='VH'))]
        with pytest.raises(ValueError, match='All bursts must have the same polarization. Found:.*'):
            Swath.check_burst_group_validity(different_polarization)

        non_consecutive_burst_ids = [burst1, BurstStub(*burst2._replace(burst_id=3))]
        with pytest.raises(ValueError, match='All bursts must have consecutive burst IDs. Found:.*'):
            Swath.check_burst_group_validity(non_consecutive_burst_ids)

    def test_get_name(self, burst_infos):
        safe_path = Path('./S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85.SAFE')
        swath = Swath(burst_infos, safe_path, 1)
        assert swath.get_name() == 's1a-iw2-slc-vv-20200604t022312-20200604t022318-032861-03ce65-001'

    def test_get_bbox(self, burst_infos):
        geo_points = [
            GeoPoint(1, 1, 0, 0, 0),
            GeoPoint(1, 2, 0, 0, 0),
            GeoPoint(1.5, 1.5, 0, 0, 0),
            GeoPoint(2, 1, 0, 0, 0),
            GeoPoint(2, 2, 0, 0, 0),
        ]
        ProductStub = namedtuple('ProductStub', ['gcps'])
        safe_path = Path('./S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85.SAFE')
        swath = Swath(burst_infos, safe_path, 1)
        swath.product = ProductStub(geo_points)
        bbox = swath.get_bbox()
        polygon = Polygon([(1, 1), (1, 2), (2, 2), (2, 1), (1, 1)])
        assert bbox == polygon



from collections import namedtuple
from copy import deepcopy
from datetime import datetime, timedelta
from pathlib import Path
from unittest.mock import patch

import lxml
import lxml.etree as ET
import pytest

from burst2safe import utils
from helpers import create_test_geotiff


def test_add_shape_info(tmp_path, burst_info1):
    tmp_tiff = tmp_path / 'tmp.tiff'
    create_test_geotiff(tmp_tiff, dtype='float', shape=(10, 10, 1))
    tmp_burst = deepcopy(burst_info1)
    tmp_burst.data_path = tmp_tiff
    tmp_burst.add_shape_info()
    assert tmp_burst.length == 10
    assert tmp_burst.width == 10


def test_add_start_stop_utc(burst_info1):
    test_elem = ET.Element('product')
    az_time_interval = ET.SubElement(test_elem, 'azimuthTimeInterval')
    az_time_interval.text = '1'
    start = datetime.fromisoformat('2020-01-01T00:00:00')
    for i in range(9):
        burst_elem = ET.SubElement(test_elem, 'burst')
        az_time = ET.SubElement(burst_elem, 'azimuthTime')
        burst_time = (start + timedelta(seconds=i * 5)).isoformat()
        az_time.text = burst_time

    with patch('burst2safe.utils.get_subxml_from_metadata', return_value=test_elem):
        tmp_burst = deepcopy(burst_info1)
        tmp_burst.burst_index = 2
        tmp_burst.length = 10
        tmp_burst.add_start_stop_utc()
        assert tmp_burst.start_utc == datetime.fromisoformat('2020-01-01T00:00:10')
        assert tmp_burst.stop_utc == datetime.fromisoformat('2020-01-01T00:00:19')


def test_create_burst_info(tmp_path, search_result1):
    burst_granule = 'S1_136231_IW2_20200604T022312_VV_7C85-BURST'
    slc_granule = 'S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85'
    burst = utils.create_burst_info(search_result1, tmp_path)

    assert burst.granule == burst_granule
    assert burst.slc_granule == slc_granule
    assert burst.swath == 'IW2'
    assert burst.polarization == 'VV'
    assert burst.burst_id == 123456
    assert burst.burst_index == 7
    assert burst.direction == 'ASCENDING'
    assert burst.absolute_orbit == 123
    assert burst.data_url == 'https://example.com/foo.zip'
    assert burst.metadata_url == 'https://example.com/foo.xml'
    assert burst.data_path == tmp_path / f'{burst_granule}.tiff'
    assert burst.metadata_path == tmp_path / f'{slc_granule}_VV.xml'


def test_get_burst_infos(burst_info1):
    with patch.object(utils, 'create_burst_info') as mock_create:
        mock_create.return_value = burst_info1
        infos = utils.get_burst_infos(['granule1', 'granule2'], Path(''))

    assert isinstance(infos, list)
    assert len(infos) == 2


def test_sort_burst_infos():
    StubInfo = namedtuple('StubInfo', ['swath', 'polarization', 'burst_id'])
    info1 = StubInfo('IW1', 'VV', 1)
    info2 = StubInfo('IW1', 'HH', 1)
    info3 = StubInfo('IW2', 'VV', 1)
    info4 = StubInfo('IW2', 'HH', 1)
    info5 = StubInfo('IW1', 'VV', 2)
    info6 = StubInfo('IW1', 'HH', 2)
    info7 = StubInfo('IW2', 'VV', 2)
    info8 = StubInfo('IW2', 'HH', 2)
    infos = [info1, info3, info5, info2, info8, info4, info6, info7]

    sorted_infos = utils.sort_burst_infos(infos)

    assert sorted_infos['IW1']['VV'] == [info1, info5]
    assert sorted_infos['IW1']['HH'] == [info2, info6]
    assert sorted_infos['IW2']['VV'] == [info3, info7]
    assert sorted_infos['IW2']['HH'] == [info4, info8]


def test_optional_wd():
    wd = utils.optional_wd()
    assert isinstance(wd, Path)
    assert wd == Path.cwd()

    existing_dir = 'working'
    wd = utils.optional_wd(existing_dir)
    assert isinstance(wd, Path)
    assert wd == Path(existing_dir)


def test_calculate_crc16(tmp_path, test_data_xml, test_data_dir):
    manifest_file = test_data_dir / 'manifest_7C85.safe'
    crc = utils.calculate_crc16(manifest_file)
    assert crc == '7C85'


@pytest.mark.parametrize('xml_type, swath', [('product', 'IW1'), ('noise', 'IW2'), ('calibration', 'IW3')])
def test_get_subxml_from_metadata(xml_type, swath, test_data_xml):
    result = utils.get_subxml_from_metadata(test_data_xml, xml_type, swath, 'VV')
    assert isinstance(result, lxml.etree._Element)
    assert result.find('adsHeader/swath').text == swath
    assert result.find('adsHeader/polarisation').text == 'VV'
    assert result.tag == 'content'


def test_get_subxml_from_metadata_invalid(test_data_xml):
    with pytest.raises(ValueError):
        utils.get_subxml_from_metadata(test_data_xml, 'invalid', 'IW1', 'VV')

    result = utils.get_subxml_from_metadata(test_data_xml, 'product', 'invalid', 'VV')
    assert result is None

    result = utils.get_subxml_from_metadata(test_data_xml, 'product', 'IW1', 'invalid')
    assert result is None


def test_get_subxml_from_metadata_manifest(test_data_xml):
    result = utils.get_subxml_from_metadata(test_data_xml, 'manifest')
    assert isinstance(result, lxml.etree._Element)
    assert result.tag == '{urn:ccsds:schema:xfdu:1}XFDU'


def test_flatten():
    assert utils.flatten([[1, 2], [3, 4], [5, 6]]) == [1, 2, 3, 4, 5, 6]


def test_drop_duplicates():
    assert utils.drop_duplicates([1, 2, 3, 4, 4, 5, 6, 6]) == [1, 2, 3, 4, 5, 6]



from burst2safe.calibration import Calibration
from helpers import validate_xml


class TestCalibration:
    def test_merge(self, burst_infos, tmp_path, xsd_dir):
        out_path = tmp_path / 'file-001.xml'
        xsd_file = xsd_dir / 's1-level-1-calibration.xsd'
        calibration = Calibration(burst_infos, 1)
        calibration.assemble()
        calibration.write(out_path)
        assert validate_xml(out_path, xsd_file)



import shutil
from itertools import product
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

import numpy as np
from shapely.geometry import MultiPolygon, Polygon

from burst2safe.manifest import Manifest
from burst2safe.swath import Swath
from burst2safe.utils import BurstInfo, drop_duplicates, get_subxml_from_metadata, optional_wd


class Safe:
    """Class representing a SAFE file."""

    def __init__(self, burst_infos: Iterable[BurstInfo], work_dir: Optional[Path] = None):
        """Initialize a Safe object.

        Args:
            burst_infos: A list of BurstInfo objects
            work_dir: The directory to create the SAFE in
        """
        self.burst_infos = burst_infos
        self.work_dir = optional_wd(work_dir)

        self.check_group_validity(self.burst_infos)

        self.grouped_burst_infos = self.group_burst_infos(self.burst_infos)
        self.name = self.get_name(self.burst_infos)
        self.safe_path = self.work_dir / self.name
        self.swaths = []
        self.manifest = None

    @staticmethod
    def check_group_validity(burst_infos: Iterable[BurstInfo]):
        """Check that the burst group is valid.

        A valid burst group must:
        - Have the same acquisition mode
        - Be from the same absolute orbit
        - Be contiguous in time and space
        - Have the same footprint for all included polarizations

        Args:
            burst_infos: A list of BurstInfo objects
        """
        swaths = sorted(list(set([info.swath for info in burst_infos])))
        polarizations = sorted(list(set([info.polarization for info in burst_infos])))
        burst_range = {}
        for swath in swaths:
            burst_range[swath] = {}
            for pol in polarizations:
                burst_subset = [info for info in burst_infos if info.swath == swath and info.polarization == pol]
                if len(burst_subset) == 0:
                    burst_range[swath][pol] = [0, 0]
                    continue
                Swath.check_burst_group_validity(burst_subset)

                burst_ids = [info.burst_id for info in burst_subset]
                burst_range[swath][pol] = [min(burst_ids), max(burst_ids)]

            start_ids = [id_range[0] for id_range in burst_range[swath].values()]
            if len(set(start_ids)) != 1:
                raise ValueError(
                    f'Polarization groups in swath {swath} do not have same start burst id. Found {start_ids}'
                )

            end_ids = [id_range[1] for id_range in burst_range[swath].values()]
            if len(set(end_ids)) != 1:
                raise ValueError(f'Polarization groups in swath {swath} do not have same end burst id. Found {end_ids}')

        if len(swaths) == 1:
            return

        swath_combos = [[swaths[i], swaths[i + 1]] for i in range(len(swaths) - 1)]
        working_pol = polarizations[0]
        for swath1, swath2 in swath_combos:
            min_diff = burst_range[swath1][working_pol][0] - burst_range[swath2][working_pol][1]
            if np.abs(min_diff) > 1:
                raise ValueError(f'Products from swaths {swath1} and {swath2} do not overlap')
            max_diff = burst_range[swath1][working_pol][1] - burst_range[swath2][working_pol][0]
            if np.abs(max_diff) > 1:
                raise ValueError(f'Products from swaths {swath1} and {swath2} do not overlap')

    @staticmethod
    def get_name(burst_infos: Iterable[BurstInfo], unique_id: str = '0000') -> str:
        """Create a name for the SAFE file.

        Args:
            burst_infos: A list of BurstInfo objects
            unique_id: A unique identifier for the SAFE file

        Returns:
            The name of the SAFE file
        """
        platform, beam_mode, product_type = burst_infos[0].slc_granule.split('_')[:3]
        product_info = f'1SS{burst_infos[0].polarization[0]}'
        min_date = min([x.date for x in burst_infos]).strftime('%Y%m%dT%H%M%S')
        max_date = max([x.date for x in burst_infos]).strftime('%Y%m%dT%H%M%S')
        absolute_orbit = f'{burst_infos[0].absolute_orbit:06d}'
        mission_data_take = burst_infos[0].slc_granule.split('_')[-2]
        product_name = f'{platform}_{beam_mode}_{product_type}__{product_info}_{min_date}_{max_date}_{absolute_orbit}_{mission_data_take}_{unique_id}.SAFE'
        return product_name

    @staticmethod
    def group_burst_infos(burst_infos: Iterable[BurstInfo]) -> dict:
        """Group burst infos by swath and polarization.

        Args:
            burst_infos: A list of BurstInfo objects

        Returns:
            A dictionary of burst infos grouped by swath, then polarization
        """
        burst_dict = {}
        for burst_info in burst_infos:
            if burst_info.swath not in burst_dict:
                burst_dict[burst_info.swath] = {}

            if burst_info.polarization not in burst_dict[burst_info.swath]:
                burst_dict[burst_info.swath][burst_info.polarization] = []

            burst_dict[burst_info.swath][burst_info.polarization].append(burst_info)

        swaths = list(burst_dict.keys())
        polarizations = list(burst_dict[swaths[0]].keys())
        for swath, polarization in zip(swaths, polarizations):
            burst_dict[swath][polarization] = sorted(burst_dict[swath][polarization], key=lambda x: x.burst_id)

        return burst_dict

    def get_bbox(self):
        """Get the bounding box for the SAFE file.

        Returns:
            A Polygon object representing the bounding box
        """
        bboxs = MultiPolygon([swath.bbox for swath in self.swaths])
        min_rotated_rect = bboxs.minimum_rotated_rectangle
        bbox = Polygon(min_rotated_rect.exterior)
        return bbox

    def create_dir_structure(self) -> Path:
        """Create a directory for the SAFE file.

        Returns:
            The path to the SAFE directory
        """
        measurements_dir = self.safe_path / 'measurement'
        annotations_dir = self.safe_path / 'annotation'
        calibration_dir = annotations_dir / 'calibration'

        calibration_dir.mkdir(parents=True, exist_ok=True)
        measurements_dir.mkdir(parents=True, exist_ok=True)

        xsd_dir = Path(__file__).parent / 'data'
        shutil.copytree(xsd_dir, self.safe_path / 'support', dirs_exist_ok=True)

    def create_safe_components(self):
        """Create the components (data and metadata files) of the SAFE file."""
        swaths = list(self.grouped_burst_infos.keys())
        polarizations = list(self.grouped_burst_infos[swaths[0]].keys())
        for i, (swath, polarization) in enumerate(product(swaths, polarizations)):
            image_number = i + 1
            burst_infos = self.grouped_burst_infos[swath][polarization]
            swath = Swath(burst_infos, self.safe_path, image_number)
            swath.assemble()
            swath.write()
            self.swaths.append(swath)

    def compile_manifest_components(self) -> Tuple[List, List, List]:
        """Compile the manifest components for all files within the SAFE file.

        Returns:
            A list of content units, metadata objects, and data objects for the manifest file
        """
        content_units = []
        metadata_objects = []
        data_objects = []
        for swath in self.swaths:
            for annotation in swath.annotations:
                content_unit, metadata_object, date_object = annotation.create_manifest_components()
                content_units.append(content_unit)
                metadata_objects.append(metadata_object)
                data_objects.append(date_object)
            measurement_content, measurement_data = swath.measurement.create_manifest_components()
            content_units.append(measurement_content)
            data_objects.append(measurement_data)
        return content_units, metadata_objects, data_objects

    def create_manifest(self):
        """Create the manifest.safe file for the SAFE file."""
        manifest_name = self.safe_path / 'manifest.safe'
        content_units, metadata_objects, data_objects = self.compile_manifest_components()
        template_manifest = get_subxml_from_metadata(self.burst_infos[0].metadata_path, 'manifest')
        manifest = Manifest(content_units, metadata_objects, data_objects, self.get_bbox(), template_manifest)
        manifest.assemble()
        manifest.write(manifest_name)
        self.manifest = manifest

    def update_product_identifier(self):
        """Update the product identifier using the CRC of the manifest file."""
        new_new = self.get_name(self.burst_infos, unique_id=self.manifest.crc)
        new_path = self.work_dir / new_new
        if new_path.exists():
            shutil.rmtree(new_path)
        shutil.move(self.safe_path, new_path)
        self.name = new_new
        self.safe_path = new_path

    def create_safe(self):
        """Create the SAFE file."""
        self.create_dir_structure()
        self.create_safe_components()
        self.create_manifest()
        self.update_product_identifier()
        return self.safe_path

    def cleanup(self):
        to_delete = [burst_info.data_path for burst_info in self.burst_infos]
        to_delete += [burst_info.metadata_path for burst_info in self.burst_infos]
        to_delete = drop_duplicates(to_delete)
        for file in to_delete:
            file.unlink()



import warnings
from binascii import crc_hqx
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, Iterable, List, Optional

import lxml.etree as ET
from asf_search.Products.S1BurstProduct import S1BurstProduct
from osgeo import gdal


gdal.UseExceptions()
warnings.filterwarnings('ignore')


@dataclass
class BurstInfo:
    """Dataclass for storing burst information."""

    granule: str
    slc_granule: str
    swath: str
    polarization: str
    burst_id: int
    burst_index: int
    direction: str
    absolute_orbit: int
    date: datetime
    data_url: Path
    data_path: Path
    metadata_url: Path
    metadata_path: Path
    start_utc: datetime = None
    stop_utc: datetime = None
    length: int = None
    width: int = None

    def add_shape_info(self):
        """Add shape information to the BurstInfo object."""
        info = gdal.Info(str(self.data_path), format='json')
        self.width, self.length = info['size']

    def add_start_stop_utc(self):
        """Add start and stop UTC to burst info.
        There is spatial overlap between bursts, so burst start/stop times will overlap as well.
        """
        annotation = get_subxml_from_metadata(self.metadata_path, 'product', self.swath, self.polarization)
        start_utcs = [datetime.fromisoformat(x.find('azimuthTime').text) for x in annotation.findall('.//burst')]
        self.start_utc = start_utcs[self.burst_index]

        azimuth_time_interval = float(annotation.find('.//azimuthTimeInterval').text)
        burst_time_interval = timedelta(seconds=(self.length - 1) * azimuth_time_interval)
        self.stop_utc = self.start_utc + burst_time_interval


def create_burst_info(product: S1BurstProduct, work_dir: Path) -> BurstInfo:
    """Create a BurstInfo object given a granule.

    Args:
        product: A S1BurstProduct object
        work_dir: The directory to save the data to.
    """
    slc_granule = product.umm['InputGranules'][0].split('-')[0]

    burst_granule = product.properties['fileID']
    direction = product.properties['flightDirection'].upper()
    polarization = product.properties['polarization'].upper()
    absolute_orbit = int(product.properties['orbit'])
    data_url = product.properties['url']
    metadata_url = product.properties['additionalUrls'][0]

    swath = product.properties['burst']['subswath'].upper()
    burst_id = int(product.properties['burst']['relativeBurstID'])
    burst_index = int(product.properties['burst']['burstIndex'])

    date_format = '%Y%m%dT%H%M%S'
    burst_time_str = burst_granule.split('_')[3]
    burst_time = datetime.strptime(burst_time_str, date_format)
    data_path = work_dir / f'{burst_granule}.tiff'
    metadata_path = work_dir / f'{slc_granule}_{polarization}.xml'

    burst_info = BurstInfo(
        burst_granule,
        slc_granule,
        swath,
        polarization,
        burst_id,
        burst_index,
        direction,
        absolute_orbit,
        burst_time,
        data_url,
        data_path,
        metadata_url,
        metadata_path,
    )
    return burst_info


def get_burst_infos(products: Iterable[S1BurstProduct], work_dir: Path) -> List[BurstInfo]:
    """Get burst information from ASF Search.

    Args:
        products: A list of S1BurstProduct objects.
        save_dir: The directory to save the data to.

    Returns:
        A list of BurstInfo objects.
    """
    work_dir = optional_wd(work_dir)
    burst_info_list = []
    for product in products:
        burst_info = create_burst_info(product, work_dir)
        burst_info_list.append(burst_info)

    return burst_info_list


def sort_burst_infos(burst_info_list: List[BurstInfo]) -> Dict:
    """Sort BurstInfo objects by swath and polarization.

    Args:
        burst_info_list: List of BurstInfo objects.

    Returns:
        Dictionary of sorted BurstInfo objects. First key is swath, second key is polarization.
    """
    burst_infos = {}
    for burst_info in burst_info_list:
        if burst_info.swath not in burst_infos:
            burst_infos[burst_info.swath] = {}

        if burst_info.polarization not in burst_infos[burst_info.swath]:
            burst_infos[burst_info.swath][burst_info.polarization] = []

        burst_infos[burst_info.swath][burst_info.polarization].append(burst_info)

    swaths = list(burst_infos.keys())
    polarizations = list(burst_infos[swaths[0]].keys())
    for swath, polarization in zip(swaths, polarizations):
        burst_infos[swath][polarization] = sorted(burst_infos[swath][polarization], key=lambda x: x.burst_id)

    return burst_infos


def optional_wd(wd: Optional[Path | str] = None) -> Path:
    """Return the working directory as a Path object

    Args:
        wd: Optional working directory as a Path or string

    Returns:
        Path to your input working directory or the current working directory.
    """
    if wd is None:
        wd = Path.cwd()
    return Path(wd)


def calculate_crc16(file_path: Path) -> str:
    """Calculate the CRC16 checksum for a file.

    Args:
        file_path: Path to file to calculate checksum for

    Returns:
        CRC16 checksum as a hexadecimal string
    """
    with open(file_path, 'rb') as f:
        data = f.read()

    crc = f'{crc_hqx(data, 0xffff):04X}'
    return crc


def get_subxml_from_metadata(
    metadata_path: Path, xml_type: str, subswath: str = None, polarization: str = None
) -> ET.Element:
    """Extract child xml info from ASF combined metadata file.

    Args:
        metadata_path: Path to metadata file
        xml_typ: Desired type of metadata to obtain (product, noise, calibration, or rfi)
        subswath: Desired subswath to obtain data for
        polarization: Desired polarization to obtain data for

    Returns:
        lxml Element for desired metadata
    """
    with open(metadata_path, 'r') as metadata_file:
        metadata = ET.parse(metadata_file).getroot()

    if xml_type == 'manifest':
        desired_metadata = metadata.find('manifest/{urn:ccsds:schema:xfdu:1}XFDU')
        return desired_metadata

    possible_types = ['product', 'noise', 'calibration', 'rfi']
    if xml_type not in possible_types:
        raise ValueError(f'Metadata type {xml_type} not one of {" ".join(possible_types)}')

    if subswath is None or polarization is None:
        raise ValueError('subswath and polarization must be provided for non-manifest files')

    correct_type = [x for x in metadata.find('metadata').iterchildren() if x.tag == xml_type]
    correct_swath = [x for x in correct_type if x.find('swath').text == subswath]
    correct_pol = [x for x in correct_swath if x.find('polarisation').text == polarization]

    if not correct_pol:
        desired_metadata = None
    else:
        desired_metadata = correct_pol[0].find('content')

    return desired_metadata


def flatten(list_of_lists: List[List]) -> List:
    """Flatten a list of lists."""
    return [item for sublist in list_of_lists for item in sublist]


def drop_duplicates(input_list: List) -> List:
    """Drop duplicates from a list, while preserving order."""
    return list(dict.fromkeys(input_list))


def set_text(element: ET.Element, text: str | int) -> None:
    """Set the text of an element if it is not None.

    Args:
        element: The element to set the text of.
        text: The text to set the element to.
    """
    if not isinstance(text, str) and not isinstance(text, int):
        raise ValueError('Text must be a string or an integer.')

    element.text = str(text)



"""A package for converting ASF burst SLCs to the SAFE format"""

import warnings
from argparse import ArgumentParser
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Iterable, List, Optional

import asf_search
from asf_search.Products.S1BurstProduct import S1BurstProduct
from shapely import box
from shapely.geometry import Polygon

from burst2safe.safe import Safe
from burst2safe.utils import BurstInfo, get_burst_infos, optional_wd


warnings.filterwarnings('ignore')

DESCRIPTION = """Convert a set of ASF burst SLCs to the ESA SAFE format.
You can either provide a list of burst granules, or define a burst group by
providing the absolute orbit number, starting burst ID, ending burst ID, swaths,
and polarizations arguments.
"""


def find_granules(granules: Iterable[str]) -> List[S1BurstProduct]:
    results = asf_search.search(product_list=granules)
    found_granules = [result.properties['fileID'] for result in results]

    for granule in granules:
        if granule not in found_granules:
            raise ValueError(f'ASF Search failed to find {granule}.')
    return list(results)


def find_group(orbit: int, footprint: Polygon, polarizations: Iterable[str]) -> List[S1BurstProduct]:
    results = []
    for pol in polarizations:
        if pol not in ['VV', 'VH', 'HV', 'HH']:
            raise ValueError(f'Invalid polarization: {pol}')

        single_pol = asf_search.geo_search(
            dataset=asf_search.constants.DATASET.SLC_BURST,
            absoluteOrbit=orbit,
            intersectsWith=footprint.wkt,
            polarization=pol,
        )

        if not single_pol:
            # TODO: add Vertex link to error message?
            raise ValueError(
                f'No results found for orbit {orbit}, footprint {footprint}, and polarization {pol}.'
                'Bursts may not populated yet for this group. Check Vertex to confirm.'
            )
        results.extend(list(single_pol))
    return results


def find_bursts(
    granules: Optional[Iterable[str]] = None,
    orbit: Optional[int] = None,
    footprint: Optional[Polygon] = None,
    polarizations: Optional[Iterable[str]] = None,
) -> Path:
    """Find bursts using ASF Search.

    Args:
        granules: A list of burst granules to convert to SAFE
        orbit: The absolute orbit number of the bursts
        bbox: The bounding box of the bursts
        polarizations: List of polarizations to include
    """
    if granules:
        print('Using granule search...')
        results = find_granules(granules)
    elif orbit and footprint and polarizations:
        print('Using burst group search...')
        results = find_group(orbit, footprint, polarizations)
    else:
        raise ValueError(
            'You must provide either a list of granules or ALL burst group parameters'
            '(Orbit, Start ID, End ID, Swaths, and Polarizations.'
        )
    return results


def download_bursts(burst_infos: Iterable[BurstInfo]):
    downloads = {}
    for burst_info in burst_infos:
        downloads[burst_info.data_path] = burst_info.data_url
        downloads[burst_info.metadata_path] = burst_info.metadata_url
    download_info = [(key.parent, key.name, value) for key, value in downloads.items()]

    dirs, names, urls = zip(*download_info)
    # TODO: bug in asf_search (issue #282) requires a new session for each download
    sessions = [asf_search.ASFSession() for _ in range(len(urls))]
    with ThreadPoolExecutor() as executor:
        executor.map(asf_search.download_url, urls, dirs, names, sessions)


def burst2safe(
    granules: Optional[Iterable[str]] = None,
    orbit: Optional[int] = None,
    footprint: Optional[Polygon] = None,
    polarizations: Optional[Iterable[str]] = None,
    keep_files: bool = False,
    work_dir: Optional[Path] = None,
) -> Path:
    """Convert a set of burst granules to the ESA SAFE format.

    To be eligible for conversions, all burst granules must:
    - Have the same acquisition mode
    - Be from the same absolute orbit
    - Be contiguous in time and space
    - Have the same footprint for all included polarizations

    Args:
        granules: A list of burst granules to convert to SAFE
        orbit: The absolute orbit number of the bursts
        footprint: The bounding box of the bursts
        polarizations: List of polarizations to include
        work_dir: The directory to create the SAFE in
    """
    work_dir = optional_wd(work_dir)

    products = find_bursts(granules, orbit, footprint, polarizations)
    burst_infos = get_burst_infos(products, work_dir)
    print(f'Found {len(burst_infos)} burst(s).')

    print('Check burst group validity...')
    Safe.check_group_validity(burst_infos)

    print('Downloading data...')
    download_bursts(burst_infos)
    print('Download complete.')

    print('Creating SAFE...')
    [info.add_shape_info() for info in burst_infos]
    [info.add_start_stop_utc() for info in burst_infos]

    safe = Safe(burst_infos, work_dir)
    safe_path = safe.create_safe()
    print('SAFE created!')

    if not keep_files:
        safe.cleanup()

    return safe_path


def main() -> None:
    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument('granules', nargs='*', help='A list of bursts to convert to SAFE')
    parser.add_argument('--orbit', type=int, help='The absolute orbit number of the bursts')
    parser.add_argument('--bbox', type=float, nargs=4, help='Bounding box of the bursts (S W N E in lat/lon)')
    parser.add_argument('--pols', type=str, nargs='+', help='The polarizations of the bursts (i.e., VV VH)')
    parser.add_argument('--keep-files', action='store_true', default=False, help='Keep the intermediate files')
    args = parser.parse_args()

    if args.bbox:
        args.bbox = box(*args.bbox)
    if args.pols:
        args.pols = [pol.upper() for pol in args.pols]

    burst2safe(
        granules=args.granules,
        orbit=args.orbit,
        footprint=args.bbox,
        polarizations=args.pols,
        keep_files=args.keep_files,
    )



from datetime import datetime, timedelta
from pathlib import Path

import lxml.etree as ET
import pytest

from burst2safe import base


METADATA_TYPE = 'product'
IMAGE_NUMBER = 1


@pytest.fixture
def annotation(burst_infos):
    return base.Annotation(burst_infos, METADATA_TYPE, IMAGE_NUMBER)


@pytest.fixture
def elem_lists():
    start_time1 = datetime.fromisoformat('2020-01-01T00:00:00')
    elem_list1 = ET.Element('fooList')

    elem_list2 = ET.Element('fooList')
    start_time2 = datetime.fromisoformat('2020-01-01T00:01:00')

    for i in range(5):
        for elem_list, start_time in zip([elem_list1, elem_list2], [start_time1, start_time2]):
            element = ET.Element('foo')
            time = (start_time + timedelta(seconds=i * 15)).isoformat()
            ET.SubElement(element, 'azimuthTime').text = time
            ET.SubElement(element, 'line').text = str(i * 100)
            elem_list.append(element)

    elem_list2.remove(elem_list2[-1])
    return [elem_list1, elem_list2]


class TestListOfListElements:
    def test_init(self, elem_lists):
        start_line = 100
        slc_lengths = [400, 300]
        list_of_lists = base.ListOfListElements(elem_lists, start_line, slc_lengths)
        assert list_of_lists.name == 'fooList'
        assert list_of_lists.subelement_name == 'foo'
        assert list_of_lists.time_field == 'azimuthTime'
        assert list_of_lists.has_line is True

    def test_get_unique_elements(self, elem_lists):
        slc_lengths = [400, 300]
        list_of_lists = base.ListOfListElements(elem_lists, slc_lengths=slc_lengths)
        remaining_elements = list_of_lists.get_unique_elements()
        assert len(remaining_elements) == 8

        lines = [x.find('line').text for x in remaining_elements]
        assert lines == ['0', '100', '200', '300', '400', '500', '600', '700']

    def test_create_time_filtered_list(self, elem_lists):
        start_line = 100
        slc_lengths = [400, 300]
        list_of_lists = base.ListOfListElements(elem_lists, start_line, slc_lengths)

        time1 = datetime.fromisoformat('2020-01-01T00:00:15')
        time2 = datetime.fromisoformat('2020-01-01T00:01:15')
        filtered = list_of_lists.create_filtered_list([time1, time2])
        assert len(filtered) == 5
        lines = [x.find('line').text for x in filtered]
        assert lines == ['0', '100', '200', '300', '400']

        filtered = list_of_lists.create_filtered_list([time1, time2], line_bounds=[200, 300])
        assert len(filtered) == 2
        lines = [x.find('line').text for x in filtered]
        assert lines == ['200', '300']


def test_create_content_unit():
    simple_name = 'test_annotation'
    unit_type = 'Metadata Unit'
    rep_id = 'S1Level1TestSchema'

    content_unit = base.create_content_unit(simple_name, unit_type, rep_id)

    assert isinstance(content_unit, ET._Element)
    assert content_unit.tag == '{urn:ccsds:schema:xfdu:1}contentUnit'
    assert content_unit.get('unitType') == unit_type
    assert content_unit.get('repID') == rep_id
    assert content_unit.findall('.//dataObjectPointer')[0].get('dataObjectID') == simple_name


def test_create_metadata_object():
    simple_name = 'test_annotation'

    metadata_object = base.create_metadata_object(simple_name)

    assert isinstance(metadata_object, ET._Element)
    assert metadata_object.tag == 'metadataObject'
    assert metadata_object.get('ID') == f'{simple_name}Annotation'
    assert metadata_object.get('classification') == 'DESCRIPTION'
    assert metadata_object.get('category') == 'DMD'
    assert metadata_object.findall('.//dataObjectPointer')[0].get('dataObjectID') == simple_name


def test_create_data_object():
    simple_name = 'test_annotation'
    relative_path = Path('./test_annotation.xml')
    rep_id = 'S1Level1TestSchema'
    mime_type = 'text/xml'
    size_bytes = 1000
    md5 = 'md5hash'

    data_object = base.create_data_object(simple_name, relative_path, rep_id, mime_type, size_bytes, md5)

    assert isinstance(data_object, ET._Element)
    assert data_object.tag == 'dataObject'
    assert data_object.get('ID') == simple_name
    assert data_object.get('repID') == rep_id

    byte_stream = data_object.find('.//byteStream')
    assert byte_stream.get('mimeType') == mime_type
    assert byte_stream.get('size') == str(size_bytes)

    file_location = byte_stream.find('.//fileLocation')
    assert file_location.get('locatorType') == 'URL'
    assert file_location.get('href') == f'./{relative_path}'

    checksum = byte_stream.find('.//checksum')
    assert checksum.get('checksumName') == 'MD5'
    assert checksum.text == md5


class TestAnnotation:
    def test_annotation_init(self, annotation, burst_infos):
        assert annotation.burst_infos == burst_infos
        assert annotation.metadata_type == METADATA_TYPE
        assert annotation.image_number == IMAGE_NUMBER
        assert annotation.metadata_paths == [burst_infos[0].metadata_path]
        assert annotation.swath == burst_infos[0].swath
        assert annotation.pol == burst_infos[0].polarization
        assert annotation.start_line == 7 * 1510
        assert annotation.stop_line == 9 * 1510

    def test_create_ads_header(self, annotation):
        assert annotation.ads_header is None
        annotation.create_ads_header()
        assert annotation.ads_header is not None
        assert annotation.ads_header.find('startTime').text is not None
        assert annotation.ads_header.find('stopTime').text is not None
        assert annotation.ads_header.find('imageNumber').text == '001'

    def test_str_method(self, annotation):
        annotation.xml = ET.Element('test')
        assert isinstance(annotation.__str__(), str)
        assert annotation.__str__().startswith('<test')

    def test_write_method(self, annotation, tmp_path):
        annotation.xml = ET.ElementTree(ET.Element('test'))

        annotation.write(tmp_path / 'test_annotation.xml', update_info=False)
        assert (tmp_path / 'test_annotation.xml').exists()
        assert annotation.size_bytes is None
        assert annotation.md5 is None

        annotation.write(tmp_path / 'test_annotation.xml', update_info=True)
        assert annotation.size_bytes > 0
        assert annotation.md5 is not None

    def test_create_manifest_components(self, annotation):
        annotation.path = Path('test.SAFE/s1-iw2-vv-annotation.xml')
        annotation.size_bytes = 1000
        annotation.md5 = 'md5hash'
        content_unit, metadata_object, data_object = annotation.create_manifest_components()

        assert isinstance(content_unit, ET._Element)
        assert content_unit.get('unitType') == 'Metadata Unit'
        assert content_unit.get('repID') == 's1Level1ProductSchema'

        assert isinstance(metadata_object, ET._Element)
        assert metadata_object.find('dataObjectPointer').get('dataObjectID') == 'products1iw2vvannotation'

        assert isinstance(data_object, ET._Element)
        assert data_object.get('ID') == 'products1iw2vvannotation'
        assert data_object.get('repID') == 's1Level1ProductSchema'
        assert data_object.find('byteStream').get('size') == '1000'
        assert data_object.find('byteStream/checksum').text == 'md5hash'



import xml.etree.ElementTree as ET

import numpy as np

from burst2safe.noise import Noise
from helpers import validate_xml


class TestNoise:
    def test_get_start_stop_indexes(self):
        lines = np.array([0, 2, 4, 6, 8, 10])

        start, stop = Noise._get_start_stop_indexes(lines, 5)
        assert start == 0
        assert stop == 3

        start, stop = Noise._get_start_stop_indexes(lines, 5, 3)
        assert start == 1
        assert stop == 3

        start, stop = Noise._get_start_stop_indexes(lines, 12)
        assert start == 0
        assert stop == 5

    def test_update_azimuth_vector(self):
        az_vector = ET.Element('azimuthVector')
        first_az_line = ET.SubElement(az_vector, 'firstAzimuthLine')
        first_az_line.text = '0'
        last_az_line = ET.SubElement(az_vector, 'lastAzimuthLine')
        last_az_line.text = '20'
        line_element = ET.SubElement(az_vector, 'line')
        line_element.text = '0 2 4 6 8 10 12 14 16 18 20'
        az_lut_element = ET.SubElement(az_vector, 'noiseAzimuthLut')
        az_lut_element.text = '0 2 4 6 8 10 12 14 16 18 20'
        for element in (line_element, az_lut_element):
            element.set('count', str(len(element.text.split(' '))))

        new_az_vector = Noise._update_azimuth_vector(az_vector, 0, 10, 20)
        assert new_az_vector.find('firstAzimuthLine').text == '0'
        assert new_az_vector.find('lastAzimuthLine').text == '10'
        assert new_az_vector.find('line').get('count') == '6'
        assert new_az_vector.find('line').text == '0 2 4 6 8 10'
        assert new_az_vector.find('noiseAzimuthLut').get('count') == '6'
        assert new_az_vector.find('noiseAzimuthLut').text == '0 2 4 6 8 10'

    def test_merge(self, burst_infos, tmp_path, xsd_dir):
        out_path = tmp_path / 'file-001.xml'
        xsd_file = xsd_dir / 's1-level-1-noise.xsd'
        noise = Noise(burst_infos, 1)
        noise.assemble()
        noise.write(out_path)
        assert validate_xml(out_path, xsd_file)



import lxml.etree as ET

from burst2safe.product import GeoPoint, Product
from helpers import validate_xml


class TestProduct:
    def test_update_data_stats(self, burst_infos):
        base_path = 'imageInformation/imageStatistics/outputData'
        product = Product(burst_infos, 1)
        product.assemble()
        product.update_data_stats(1 + 2j, 3 + 4j)
        for elem in [product.image_annotation, product.xml.find('imageAnnotation')]:
            elem.find(f'{base_path}Mean/re').text = '1'
            elem.find(f'{base_path}Mean/im').text = '2'
            elem.find(f'{base_path}StdDev/re').text = '3'
            elem.find(f'{base_path}StdDev/im').text = '4'

    def test_update_gcps(self, burst_infos):
        geolocation_grid = ET.Element('geolocationGrid')
        grid_point_list = ET.SubElement(geolocation_grid, 'geolocationGridPointList')
        grid_point = ET.SubElement(grid_point_list, 'geolocationGridPoint')
        lon = ET.SubElement(grid_point, 'longitude')
        lon.text = '1.1'
        lat = ET.SubElement(grid_point, 'latitude')
        lat.text = '2.2'
        hgt = ET.SubElement(grid_point, 'height')
        hgt.text = '3.3'
        line = ET.SubElement(grid_point, 'line')
        line.text = '4'
        pixel = ET.SubElement(grid_point, 'pixel')
        pixel.text = '5'

        product = Product(burst_infos, 1)
        product.geolocation_grid = geolocation_grid

        assert len(product.gcps) == 0

        product.update_gcps()
        assert len(product.gcps) == 1
        assert product.gcps[0] == GeoPoint(1.1, 2.2, 3.3, 4, 5)

    def test_merge(self, burst_infos, tmp_path, xsd_dir):
        out_path = tmp_path / 'file-001.xml'
        xsd_file = xsd_dir / 's1-level-1-product.xsd'
        product = Product(burst_infos, 1)
        product.assemble()

        # Add back in omitted fields
        product.update_data_stats(1 + 1j, 2 + 2j)
        burst_elems = product.xml.findall('.//byteOffset')
        for burst_elem in burst_elems:
            burst_elem.text = '1'
        product.xml.find('generalAnnotation/productInformation/platformHeading').text = '0.1'

        product.write(out_path)
        assert validate_xml(out_path, xsd_file)



from copy import deepcopy
from typing import Iterable

import lxml.etree as ET
import numpy as np

from burst2safe.base import Annotation
from burst2safe.utils import BurstInfo, flatten


class Noise(Annotation):
    """Class representing a Noise XML."""

    def __init__(self, burst_infos: Iterable[BurstInfo], image_number: int):
        """Create a Noise object.

        Args:
            burst_infos: List of BurstInfo objects.
            image_number: Image number.
        """
        super().__init__(burst_infos, 'noise', image_number)
        self.range_vector_list = None
        self.azimuth_vector_list = None

    def create_range_vector_list(self):
        """Create the range vector list."""
        self.range_vector_list = self.merge_lists('noiseRangeVectorList')

    @staticmethod
    def _get_start_stop_indexes(lines: np.ndarray, last_line: int, first_line: int = 0) -> tuple[int, int]:
        """Get the indexes of the first and last lines in the range of lines.

        Args:
            lines: Array of lines.
            last_line: Last line of the range.
            first_line: First line of the range. Defaults to 0.

        Returns:
            Tuple of the indexes of the first and last lines in the range.
        """
        if np.any(lines <= first_line):
            first_index = np.where(lines == lines[lines <= first_line].max())[0][0]
        else:
            first_index = 0

        if np.any(lines >= last_line):
            last_index = np.where(lines == lines[lines >= last_line].min())[0][0]
        else:
            last_index = lines.shape[0] - 1

        return first_index, last_index

    @staticmethod
    def _update_azimuth_vector(az_vector: ET.Element, line_offset: int, start_line: int, stop_line: int) -> ET.Element:
        """Update the azimuth vector to match the new line range. Subset noiseAzimuthLut to match.

        Args:
            az_vector: Azimuth vector.
            line_offset: Line offset.
            start_line: Start line.
            stop_line: Stop line.

        Returns:
            Updated azimuth vector.
        """
        new_az_vector = deepcopy(az_vector)

        line_element = new_az_vector.find('line')
        lines = np.array([int(x) for x in line_element.text.split(' ')])
        lines += line_offset

        first_index, last_index = Noise._get_start_stop_indexes(lines, stop_line - start_line - 1)
        slice = np.s_[first_index : last_index + 1]
        count = str(last_index - first_index + 1)

        new_az_vector.find('firstAzimuthLine').text = str(lines[first_index])
        new_az_vector.find('lastAzimuthLine').text = str(lines[last_index])

        line_element.text = ' '.join([str(x) for x in lines[slice]])
        line_element.set('count', count)

        az_lut_element = new_az_vector.find('noiseAzimuthLut')
        az_lut_element.text = ' '.join(az_lut_element.text.split(' ')[slice])
        az_lut_element.set('count', count)
        return new_az_vector

    def create_azimuth_vector_list(self):
        """Create the azimuth vector list. ListOfListElements class can't be used here because the
        noiseAzimuthVectorList has a different structure than the other lists elements.
        """
        az_vectors = [noise.find('noiseAzimuthVectorList') for noise in self.inputs]
        updated_az_vectors = []
        for i, az_vector_set in enumerate(az_vectors):
            slc_offset = sum(self.slc_lengths[:i])
            az_vectors = az_vector_set.findall('noiseAzimuthVector')
            updated_az_vector_set = []
            for az_vector in az_vectors:
                line_offset = slc_offset - self.start_line
                updated_az_vector = self._update_azimuth_vector(az_vector, line_offset, self.start_line, self.stop_line)
                updated_az_vector_set.append(updated_az_vector)
            updated_az_vectors.append(updated_az_vector_set)

        updated_az_vectors = flatten(updated_az_vectors)

        new_az_vector_list = ET.Element('noiseAzimuthVectorList')
        new_az_vector_list.set('count', str(len(updated_az_vectors)))
        for az_vector in updated_az_vectors:
            new_az_vector_list.append(az_vector)
        self.azimuth_vector_list = new_az_vector_list

    def assemble(self):
        """Assemble the Noise object from its components."""
        self.create_ads_header()
        self.create_range_vector_list()
        self.create_azimuth_vector_list()

        noise = ET.Element('noise')
        noise.append(self.ads_header)
        noise.append(self.range_vector_list)
        noise.append(self.azimuth_vector_list)
        noise_tree = ET.ElementTree(noise)

        ET.indent(noise_tree, space='  ')
        self.xml = noise_tree



import hashlib
from datetime import datetime
from pathlib import Path
from typing import Iterable, Tuple

import numpy as np
from osgeo import gdal, osr

from burst2safe.base import create_content_unit, create_data_object
from burst2safe.product import GeoPoint
from burst2safe.utils import BurstInfo, get_subxml_from_metadata


class Measurement:
    """Class representing a measurement GeoTIFF."""

    def __init__(self, burst_infos: Iterable[BurstInfo], gcps: Iterable[GeoPoint], image_number: int):
        """Initialize a Measurement object.

        Args:
            burst_infos: A list of BurstInfo objects
            gcps: A list of GeoPoint objects
            image_number: The image number of the measurement
        """
        self.burst_infos = burst_infos
        self.gcps = gcps
        self.image_number = image_number

        self.swath = self.burst_infos[0].swath
        self.burst_length = self.burst_infos[0].length
        self.burst_width = self.burst_infos[0].width
        self.total_width = self.burst_width
        self.total_length = self.burst_length * len(self.burst_infos)
        self.data_mean = None
        self.data_std = None
        self.size_bytes = None
        self.md5 = None

    def get_data(self, band: int = 1) -> np.ndarray:
        """Get the data from the measurement from ASF burst GeoTIFFs.

        Args:
            band: The GeoTIFF band to read

        Returns:
            The data from burst GeoTIFFs as a numpy array
        """
        data = np.zeros((self.total_length, self.total_width), dtype=np.complex64)
        for i, burst_info in enumerate(self.burst_infos):
            ds = gdal.Open(str(burst_info.data_path))
            data[i * self.burst_length : (i + 1) * self.burst_length, :] = ds.GetRasterBand(band).ReadAsArray()
            ds = None
        return data

    @staticmethod
    def get_ipf_version(metadata_path: Path) -> str:
        """Get the IPF version from the parent manifest file.

        Returns:
            The IPF version as a string
        """
        manifest = get_subxml_from_metadata(metadata_path, 'manifest')
        version_xml = [elem for elem in manifest.findall('.//{*}software') if elem.get('name') == 'Sentinel-1 IPF'][0]
        return version_xml.get('version')

    def add_metadata(self, dataset: gdal.Dataset):
        """Add metadata to an existing GDAL dataset.

        Args:
            dataset: The GDAL dataset to add metadata to
        """
        gdal_gcps = [gdal.GCP(gcp.x, gcp.y, gcp.z, gcp.pixel, gcp.line) for gcp in self.gcps]
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(4326)
        dataset.SetGCPs(gdal_gcps, srs.ExportToWkt())

        dataset.SetMetadataItem('TIFFTAG_DATETIME', datetime.strftime(datetime.now(), '%Y:%m:%d %H:%M:%S'))
        # TODO make sure A/B is being set correctly.
        dataset.SetMetadataItem('TIFFTAG_IMAGEDESCRIPTION', 'Sentinel-1A IW SLC L1')

        version = self.get_ipf_version(self.burst_infos[0].metadata_path)
        software_version = f'Sentinel-1 IPF {version}'
        dataset.SetMetadataItem('TIFFTAG_SOFTWARE', software_version)

    def create_geotiff(self, out_path: Path, update_info=True):
        """Create a GeoTIFF of SLC data from the constituent burst SLC GeoTIFFs.
        Optionally update Measurment metadata.

        Args:
            out_path: The path to write the SLC GeoTIFF to
            update_info: Whether to update the Measurement metadata
        """
        mem_drv = gdal.GetDriverByName('MEM')
        mem_ds = mem_drv.Create('', self.total_width, self.total_length, 1, gdal.GDT_CInt16)
        data = self.get_data()
        band = mem_ds.GetRasterBand(1)
        band.WriteArray(data)
        band.SetNoDataValue(0)
        self.add_metadata(mem_ds)
        gdal.Translate(str(out_path), mem_ds, format='GTiff')
        mem_ds = None
        self.path = out_path
        if update_info:
            self.data_mean = np.mean(data[data != 0])
            self.data_std = np.std(data[data != 0])
            with open(self.path, 'rb') as f:
                file_bytes = f.read()
                self.size_bytes = len(file_bytes)
                self.md5 = hashlib.md5(file_bytes).hexdigest()

    def create_manifest_components(self) -> Tuple:
        """Create the components of the SAFE manifest for the measurement file.

        Returns:
            A tuple of the content unit and data object for the measurement file
        """
        simple_name = self.path.with_suffix('').name.replace('-', '')
        rep_id = 's1Level1MeasurementSchema'
        unit_type = 'Measurement Data Unit'
        mime_type = 'application/octet-stream'

        safe_index = [i for i, x in enumerate(self.path.parts) if 'SAFE' in x][-1]
        safe_path = Path(*self.path.parts[: safe_index + 1])
        relative_path = self.path.relative_to(safe_path)

        content_unit = create_content_unit(simple_name, rep_id, unit_type)
        data_object = create_data_object(simple_name, relative_path, rep_id, mime_type, self.size_bytes, self.md5)
        return content_unit, data_object

    def write(self, out_path):
        """Write the measurement GeoTIFF to a file."""
        self.create_geotiff(out_path)



import lxml.etree as ET
import pytest
from shapely.geometry import Polygon

from burst2safe.manifest import Manifest
from burst2safe.utils import get_subxml_from_metadata


@pytest.fixture
def manifest(test_data_xml):
    content_units = [ET.Element('content_unit')]
    metadata_objects = [ET.Element('metadata_object')]
    data_objects = [ET.Element('data_object')]
    bbox = Polygon([(1, 1), (1, 2), (2, 2), (2, 1), (1, 1)])
    template_manifest = get_subxml_from_metadata(test_data_xml, 'manifest')
    test_manifest = Manifest(content_units, metadata_objects, data_objects, bbox, template_manifest)
    return test_manifest


class TestManifest:
    def test_init(self, manifest):
        assert len(manifest.content_units) == 1
        assert len(manifest.metadata_objects) == 1
        assert len(manifest.data_objects) == 1

        assert manifest.information_package_map is None
        assert manifest.metadata_section is None
        assert manifest.data_object_section is None

    def test_create_information_package_map(self, manifest):
        manifest.create_information_package_map()
        assert manifest.information_package_map is not None
        assert manifest.information_package_map.tag == '{urn:ccsds:schema:xfdu:1}informationPackageMap'
        assert len(manifest.information_package_map.find('{urn:ccsds:schema:xfdu:1}contentUnit')) == 1

    def test_create_metadata_section(self, manifest):
        manifest.create_metadata_section()
        assert manifest.metadata_section is not None
        assert manifest.metadata_section.tag == 'metadataSection'
        assert manifest.metadata_section.find('.//{*}coordinates').text == '2.0,2.0 1.0,2.0 1.0,1.0 2.0,1.0'

    def test_create_data_object_section(self, manifest):
        manifest.create_data_object_section()
        assert manifest.data_object_section is not None
        assert manifest.data_object_section.tag == 'dataObjectSection'
        assert len(manifest.data_object_section) == 1

    def test_assemble(self, manifest):
        manifest.assemble()
        assert isinstance(manifest.xml, ET._ElementTree)
        assert manifest.xml.getroot().tag == '{urn:ccsds:schema:xfdu:1}XFDU'
        assert manifest.xml.getroot().get('version') == 'esa/safe/sentinel-1.0/sentinel-1/sar/level-1/slc/standard/iwdp'
        assert manifest.xml.find('{*}informationPackageMap') is not None
        assert manifest.xml.find('metadataSection') is not None
        assert manifest.xml.find('dataObjectSection') is not None

    def test_write(self, manifest, tmp_path):
        manifest_path = tmp_path / 'manifest.safe'
        manifest.assemble()
        manifest.write(manifest_path)
        assert manifest_path.exists()
        assert manifest.path == manifest_path
        assert manifest.crc is not None



from copy import deepcopy
from dataclasses import dataclass
from datetime import timedelta
from typing import Iterable

import lxml.etree as ET
import numpy as np

from burst2safe.base import Annotation, ListOfListElements
from burst2safe.utils import BurstInfo, flatten


@dataclass
class GeoPoint:
    """A geolocation grid point."""

    x: float
    y: float
    z: float
    line: int
    pixel: int


class Product(Annotation):
    """Class representing a product XML."""

    def __init__(self, burst_infos: Iterable[BurstInfo], image_number: int):
        """Create a Product object.

        Args:
            burst_infos: A list of BurstInfo objects
            image_number: The image number
        """
        super().__init__(burst_infos, 'product', image_number)
        self.qulatity_information = None
        self.general_annotation = None
        self.image_annotation = None
        self.doppler_centroid = None
        self.antenna_pattern = None
        self.swath_timing = None
        self.geolocation_grid = None
        self.coordinate_conversion = None
        self.swath_merging = None
        self.gcps = []

    def create_quality_information(self):
        """Create the qualityInformation element."""
        quality_information = ET.Element('qualityInformation')
        quality_information.append(deepcopy(self.inputs[0].find('qualityInformation/productQualityIndex')))

        quality_datas = flatten([cal.findall('qualityInformation/qualityDataList/qualityData') for cal in self.inputs])
        quality_data_list = ET.Element('qualityDataList')
        quality_data_list.set('count', str(len(quality_datas)))
        for quality_data in quality_datas:
            quality_data_list.append(deepcopy(quality_data))

        quality_information.append(quality_data_list)

        self.quality_information = quality_information

    def create_general_annotation(self):
        """Create the generalAnnotation element.

        From product specification:
        The productInformation sub-record contains single value fields that
        are merged and included. All other sub-records contain lists which are
        concatenated. Details are presented in Table 3-11."""
        general_annotation = ET.Element('generalAnnotation')

        product_information = deepcopy(self.inputs[0].find('generalAnnotation/productInformation'))
        # TODO: productInformation/platformHeading should be calculated more accurately
        product_information.find('platformHeading').text = ''
        general_annotation.append(product_information)

        lists = [
            'downlinkInformationList',
            'orbitList',
            'attitudeList',
            'rawDataAnalysisList',
            'replicaInformationList',
            'noiseList',
            'terrainHeightList',
            'azimuthFmRateList',
        ]
        for list_name in lists:
            list_elements = [prod.find(f'generalAnnotation/{list_name}') for prod in self.inputs]
            if list_name == 'replicaInformationList':
                lol = ListOfListElements(list_elements, self.start_line, self.slc_lengths)
                unique = lol.get_unique_elements()
                filtered = ET.Element('replicaInformationList')
                filtered.set('count', str(len(unique)))
                [filtered.append(element) for element in unique]
            else:
                lol = ListOfListElements(list_elements, self.start_line, self.slc_lengths)
                filtered = lol.create_filtered_list([self.min_anx, self.max_anx], buffer=timedelta(seconds=500))

            general_annotation.append(filtered)

        self.general_annotation = general_annotation

    def create_image_annotation(self):
        """Create the imageAnnotation element.

        From product specification:
        This DSR contains two records which contain only single value fields.
        The fields in the imageInformation record are included and merged
        and all the fields for the processingInformation record are included;
        except for the inputDimensionsList record, which is concatenated.
        Details are presented in Table 3-12."""
        image_annotation = ET.Element('imageAnnotation')

        image_information = deepcopy(self.inputs[0].find('imageAnnotation/imageInformation'))
        image_information.find('productFirstLineUtcTime').text = self.min_anx.isoformat()
        image_information.find('productLastLineUtcTime').text = self.max_anx.isoformat()
        image_information.find('productComposition').text = 'Assembled'
        image_information.find('sliceNumber').text = '0'

        slice_list = image_information.find('sliceList')
        slice_list.set('count', '0')
        for element in slice_list:
            slice_list.remove(element)

        image_information.find('numberOfLines').text = str(self.total_lines)

        az_spacing_path = 'imageAnnotation/imageInformation/azimuthPixelSpacing'
        az_spacing = np.mean([float(prod.find(az_spacing_path).text) for prod in self.inputs])
        image_information.find('azimuthPixelSpacing').text = f'{az_spacing:.6e}'

        image_information.find('imageStatistics/outputDataMean/re').text = ''
        image_information.find('imageStatistics/outputDataMean/im').text = ''
        image_information.find('imageStatistics/outputDataStdDev/re').text = ''
        image_information.find('imageStatistics/outputDataStdDev/im').text = ''

        image_annotation.append(image_information)

        processing_information = deepcopy(self.inputs[0].find('imageAnnotation/processingInformation'))
        dimensions_list = processing_information.find('inputDimensionsList')
        for element in slice_list:
            dimensions_list.remove(element)

        list_elements = [prod.find('imageAnnotation/processingInformation/inputDimensionsList') for prod in self.inputs]
        lol = ListOfListElements(list_elements, self.start_line, self.slc_lengths)
        filtered = lol.create_filtered_list([self.min_anx, self.max_anx])
        [dimensions_list.append(element) for element in filtered]

        image_annotation.append(processing_information)
        self.image_annotation = image_annotation

    def update_data_stats(self, data_mean: np.complex64, data_std: np.complex64):
        """Update the data statistics in the imageAnnotation element.

        Args:
            data_mean: The complex mean of the data.
            data_std: The complex standard deviation of the data.
        """
        base_path = 'imageInformation/imageStatistics/outputData'
        data_mean_re = f'{data_mean.real:.6e}'
        data_mean_im = f'{data_mean.imag:.6e}'
        data_std_re = f'{data_std.real:.6e}'
        data_std_im = f'{data_std.imag:.6e}'

        for elem in [self.image_annotation, self.xml.find('imageAnnotation')]:
            elem.find(f'{base_path}Mean/re').text = data_mean_re
            elem.find(f'{base_path}Mean/im').text = data_mean_im
            elem.find(f'{base_path}StdDev/re').text = data_std_re
            elem.find(f'{base_path}StdDev/im').text = data_std_im

    def create_doppler_centroid(self):
        """Create the dopplerCentroid element."""
        doppler_centroid = ET.Element('dopplerCentroid')
        doppler_centroid.append(self.merge_lists('dopplerCentroid/dcEstimateList'))
        self.doppler_centroid = doppler_centroid

    def create_antenna_pattern(self):
        """Create the antennaPattern element."""
        antenna_pattern = ET.Element('antennaPattern')
        antenna_pattern.append(self.merge_lists('antennaPattern/antennaPatternList'))
        self.antenna_pattern = antenna_pattern

    def create_swath_timing(self):
        """Create the swathTiming element."""
        burst_lists = [prod.find('swathTiming/burstList') for prod in self.inputs]
        burst_lol = ListOfListElements(burst_lists, self.start_line, self.slc_lengths)
        filtered = burst_lol.create_filtered_list([self.min_anx, self.max_anx], buffer=timedelta(seconds=0.1))

        # TODO: This is needed since we always buffer backward AND forward
        if int(filtered.get('count')) > len(self.burst_infos):
            filtered.remove(filtered[-1])
            filtered.set('count', str(int(filtered.get('count')) - 1))

        # TODO: need to update burst byteOffset field
        for burst in filtered:
            burst.find('byteOffset').text = ''

        swath_timing = ET.Element('swathTiming')
        swath_timing.append(deepcopy(self.inputs[0].find('swathTiming/linesPerBurst')))
        swath_timing.append(deepcopy(self.inputs[0].find('swathTiming/samplesPerBurst')))
        swath_timing.append(filtered)
        self.swath_timing = swath_timing

    def update_gcps(self):
        """Update gcp attribute using the geolocationGridPointList."""
        gcp_xmls = self.geolocation_grid.find('geolocationGridPointList').findall('*')
        for gcp_xml in gcp_xmls:
            gcp = GeoPoint(
                float(gcp_xml.find('longitude').text),
                float(gcp_xml.find('latitude').text),
                float(gcp_xml.find('height').text),
                int(gcp_xml.find('line').text),
                int(gcp_xml.find('pixel').text),
            )
            self.gcps.append(gcp)

    def create_geolocation_grid(self):
        """Create the geolocationGrid element."""
        geolocation_grid = ET.Element('geolocationGrid')
        grid_list = self.merge_lists('geolocationGrid/geolocationGridPointList', line_bounds=[0, self.total_lines])
        geolocation_grid.append(grid_list)
        self.geolocation_grid = geolocation_grid
        self.update_gcps()

    def create_coordinate_conversion(self):
        """Create an empty coordinateConversion element."""
        coordinate_conversion = ET.Element('coordinateConversion')
        coordinate_conversion_list = ET.SubElement(coordinate_conversion, 'coordinateConversionList')
        coordinate_conversion_list.set('count', '0')
        self.coordinate_conversion = coordinate_conversion

    def create_swath_merging(self):
        """Create an empty swathMerging element."""
        swath_merging = ET.Element('swathMerging')
        swath_merge_list = ET.SubElement(swath_merging, 'swathMergeList')
        swath_merge_list.set('count', '0')
        self.swath_merging = swath_merging

    def assemble(self):
        """Assemble the product from its components."""
        self.create_ads_header()
        self.create_quality_information()
        self.create_general_annotation()
        self.create_image_annotation()
        self.create_doppler_centroid()
        self.create_antenna_pattern()
        self.create_swath_timing()
        self.create_geolocation_grid()
        self.create_coordinate_conversion()
        self.create_swath_merging()

        product = ET.Element('product')
        product.append(self.ads_header)
        product.append(self.quality_information)
        product.append(self.general_annotation)
        product.append(self.image_annotation)
        product.append(self.doppler_centroid)
        product.append(self.antenna_pattern)
        product.append(self.swath_timing)
        product.append(self.geolocation_grid)
        product.append(self.coordinate_conversion)
        product.append(self.swath_merging)
        product_tree = ET.ElementTree(product)

        ET.indent(product_tree, space='  ')
        self.xml = product_tree



from copy import deepcopy

import numpy as np
import pytest
from osgeo import gdal

from burst2safe.measurement import Measurement
from burst2safe.product import GeoPoint
from helpers import create_test_geotiff


gdal.UseExceptions()


@pytest.fixture
def burst_datas(burst_infos, tmp_path):
    burst_datas = []
    for i, burst_info in enumerate(burst_infos):
        n = i + 1
        burst_data = deepcopy(burst_info)
        burst_data.data_path = tmp_path / f'burst{n}.tif'
        burst_data.length = 10
        burst_data.width = 20

        shape = (burst_data.length, burst_data.width, 1)
        create_test_geotiff(burst_data.data_path, dtype='float', value=n, shape=shape)
        burst_datas.append(burst_data)
    return burst_datas


@pytest.fixture
def gcps():
    gcps = [GeoPoint(100, 200, 300, 1, 2), GeoPoint(400, 500, 600, 3, 4)]
    return gcps


class TestMeasurement:
    def test_init(self, burst_datas, gcps):
        measurement = Measurement(burst_datas, gcps, 1)

        assert measurement.total_length == 10 * 2
        assert measurement.data_mean is None
        assert measurement.data_std is None

    def test_get_data(self, burst_datas, gcps):
        measurement = Measurement(burst_datas, gcps, 1)
        data = measurement.get_data()
        assert data.shape == (10 * 2, 20)

        golden = np.ones((20, 20), dtype=np.complex64)
        golden[10:, :] *= 2
        assert np.allclose(data, golden)

    def test_get_ipf_version(self, burst_datas, gcps, tmp_path):
        version = Measurement.get_ipf_version(burst_datas[0].metadata_path)
        assert version == '003.20'

    def test_add_metadata(self, burst_datas, gcps):
        mem_drv = gdal.GetDriverByName('MEM')
        mem_ds = mem_drv.Create('', 20, 20, 1, gdal.GDT_CInt16)

        measurement = Measurement(burst_datas, gcps, 1)
        measurement.add_metadata(mem_ds)

        assert mem_ds.GetGCPCount() == len(gcps)
        assert '4326' in mem_ds.GetGCPProjection()
        assert mem_ds.GetMetadataItem('TIFFTAG_DATETIME') is not None
        assert mem_ds.GetMetadataItem('TIFFTAG_IMAGEDESCRIPTION') == 'Sentinel-1A IW SLC L1'
        assert mem_ds.GetMetadataItem('TIFFTAG_SOFTWARE') == 'Sentinel-1 IPF 003.20'

    def test_create_geotiff(self, burst_datas, gcps, tmp_path):
        out_path = tmp_path / 'test.tif'
        measurement = Measurement(burst_datas, gcps, 1)
        measurement.create_geotiff(out_path)

        assert out_path.exists()
        assert measurement.path == out_path
        assert measurement.data_mean is not None
        assert measurement.data_std is not None
        assert measurement.size_bytes is not None
        assert measurement.md5 is not None

    def test_create_manifest_components(self, burst_datas, gcps, tmp_path):
        measurement = Measurement(burst_datas, gcps, 1)
        measurement.path = tmp_path / 'foo.SAFE' / 'test.tif'
        measurement.size_bytes = 100
        measurement.md5 = 'md5'

        content_unit, data_unit = measurement.create_manifest_components()

        assert content_unit.get('repID') == 'Measurement Data Unit'
        assert content_unit.get('unitType') == 's1Level1MeasurementSchema'

        assert data_unit.find('byteStream').get('mimeType') == 'application/octet-stream'
        assert data_unit.find('byteStream').get('size') == '100'
        assert data_unit.find('byteStream/fileLocation').get('href') == './test.tif'
        assert data_unit.find('byteStream/checksum').text == 'md5'



import hashlib
from copy import deepcopy
from datetime import datetime, timedelta
from pathlib import Path
from typing import Iterable, List, Optional

import lxml.etree as ET

from burst2safe.utils import BurstInfo, drop_duplicates, flatten, get_subxml_from_metadata, set_text


SCHEMA = '{urn:ccsds:schema:xfdu:1}'


class ListOfListElements:
    def __init__(
        self, inputs: List[ET.Element], start_line: Optional[int] = None, slc_lengths: Optional[List[int]] = None
    ):
        """Initialize the ListOfListElements object.

        Args:
            inputs: The list of elements to be processed.
            start_line: The starting line number of the first element.
            slc_lengths: The total line lengths of the SLCs corresponding to each element.
        """
        self.inputs = inputs
        self.start_line = start_line
        self.slc_lengths = slc_lengths

        self.name = self.inputs[0].tag
        elements = flatten([element.findall('*') for element in self.inputs])
        names = drop_duplicates([x.tag for x in elements])
        if len(names) != 1:
            raise ValueError('Elements must contain only one type of subelement.')
        self.subelement_name = names[0]

        if self.name == 'replicaInformationList':
            self.time_field = 'referenceReplica/azimuthTime'
        elif 'azimuthTime' in [x.tag for x in elements[0].iter()]:
            self.time_field = 'azimuthTime'
        elif 'time' in [x.tag for x in elements[0].iter()]:
            self.time_field = 'time'
        else:
            raise ValueError('Time field not found in elements.')

        self.inputs = sorted(self.inputs, key=self.get_first_time)
        self.has_line = elements[0].find('line') is not None

    def get_first_time(self, element: ET.Element) -> datetime:
        """Get first time in List element

        args:
            element: The element to get the time from.

        Returns:
            The first time in the element.
        """
        first_time = min([datetime.fromisoformat(sub.find(self.time_field).text) for sub in element])
        return first_time

    def get_unique_elements(self) -> List[ET.Element]:
        """Get the elements without duplicates. Adjust line number if present.

        Returns:
            The list of elements without duplicates.
        """
        list_of_element_lists = [item.findall('*') for item in self.inputs]

        last_time = datetime.fromisoformat(list_of_element_lists[0][-1].find(self.time_field).text)
        uniques = [deepcopy(element) for element in list_of_element_lists[0]]
        if self.has_line:
            previous_line_count = self.slc_lengths[0]

        for i, element_list in enumerate(list_of_element_lists[1:]):
            times = [datetime.fromisoformat(element.find(self.time_field).text) for element in element_list]
            keep_index = [index for index, time in enumerate(times) if time > last_time]
            to_keep = [deepcopy(element_list[index]) for index in keep_index]

            if self.has_line:
                new_lines = [int(elem.find('line').text) + previous_line_count for elem in to_keep]
                [set_text(elem.find('line'), line) for elem, line in zip(to_keep, new_lines)]
                previous_line_count += self.slc_lengths[i]

            last_time = max([times[index] for index in keep_index])
            uniques += to_keep

        return uniques

    @staticmethod
    def filter_by_line(element_list: List[ET.Element], line_bounds: tuple[float, float]) -> List[ET.Element]:
        """Filter elements by line number.

        Args:
            line_bounds: The bounds of the line numbers.

        Returns:
            A filtered element list.
        """

        new_list = []
        for elem in element_list:
            if line_bounds[0] <= int(elem.find('line').text) <= line_bounds[1]:
                new_list.append(deepcopy(elem))
        return new_list

    def update_line_numbers(self, elements: List[ET.Element]) -> None:
        """Update the line numbers of the elements.

        Args:
            elements: The list of elements to update.
        """
        for element in elements:
            standard_line = int(element.find('line').text)
            element.find('line').text = str(standard_line - self.start_line)

    def filter_by_time(
        self, elements: List[ET.Element], anx_bounds: tuple[float, float], buffer: timedelta
    ) -> List[ET.Element]:
        """Filter elements by time.

        Args:
            elements: The list of elements to filter.
            anx_bounds: The bounds of the ANX time.
            buffer: The buffer to add to the ANX bounds.

        Returns:
            A filtered element list.
        """
        min_anx_bound = anx_bounds[0] - buffer
        max_anx_bound = anx_bounds[1] + buffer
        filtered_elements = []
        for element in elements:
            azimuth_time = datetime.fromisoformat(element.find(self.time_field).text)
            if min_anx_bound < azimuth_time < max_anx_bound:
                filtered_elements.append(deepcopy(element))

        return filtered_elements

    def create_filtered_list(
        self,
        anx_bounds: Optional[tuple[float, float]],
        buffer: Optional[timedelta] = timedelta(seconds=3),
        line_bounds: Optional[tuple[float, float]] = None,
    ) -> ET.Element:
        """Filter elements by time/line. Adjust line number if present.

        Args:
            anx_bounds: The bounds of the ANX time.
            buffer: The buffer to add to the ANX bounds.
            line_bounds: The bounds of the line numbers.

        Returns:
            A filtered list element.
        """
        elements = self.get_unique_elements()
        filtered_elements = self.filter_by_time(elements, anx_bounds, buffer)

        if self.has_line:
            self.update_line_numbers(filtered_elements)

        if line_bounds is not None:
            if not self.has_line:
                raise ValueError('Line bounds cannot be applied to elements without line numbers.')
            filtered_elements = self.filter_by_line(filtered_elements, line_bounds)

        new_element = ET.Element(self.name)
        [new_element.append(element) for element in filtered_elements]
        new_element.set('count', str(len(filtered_elements)))
        return new_element


def create_content_unit(simple_name: str, unit_type: str, rep_id: str) -> ET.Element:
    """Create a content unit element for a manifest.safe file.

    Args:
        simple_name: The simple name of the content unit.
        unit_type: The type of the content unit.
        rep_id: The representation ID.

    Returns:
        The content unit element.
    """
    content_unit = ET.Element(f'{SCHEMA}contentUnit')
    content_unit.set('unitType', unit_type)
    content_unit.set('repID', rep_id)
    ET.SubElement(content_unit, 'dataObjectPointer', dataObjectID=simple_name)
    return content_unit


def create_metadata_object(simple_name: str) -> ET.Element:
    """Create a metadata object element for a manifest.safe file.

    Args:
        simple_name: The simple name of the metadata object.

    Returns:
        The metadata object element.
    """
    metadata_object = ET.Element('metadataObject')
    metadata_object.set('ID', f'{simple_name}Annotation')
    metadata_object.set('classification', 'DESCRIPTION')
    metadata_object.set('category', 'DMD')
    ET.SubElement(metadata_object, 'dataObjectPointer', dataObjectID=simple_name)
    return metadata_object


def create_data_object(
    simple_name: str, relative_path: Path, rep_id: str, mime_type: str, size_bytes: int, md5: str
) -> ET.Element:
    """Create a data object element for a manifest.safe file.

    Args:
        simple_name: The simple name of the data object.
        relative_path: The relative path to the data object.
        rep_id: The representation ID.
        mime_type: The MIME type of the data object.
        size_bytes: The size of the data object in bytes.
        md5: The MD5 checksum of the data object.

    Returns:
        The data object element.
    """
    data_object = ET.Element('dataObject')
    data_object.set('ID', simple_name)
    data_object.set('repID', rep_id)
    byte_stream = ET.SubElement(data_object, 'byteStream')
    byte_stream.set('mimeType', mime_type)
    byte_stream.set('size', str(size_bytes))
    file_location = ET.SubElement(byte_stream, 'fileLocation')
    file_location.set('locatorType', 'URL')
    file_location.set('href', f'./{relative_path}')
    checksum = ET.SubElement(byte_stream, 'checksum')
    checksum.set('checksumName', 'MD5')
    checksum.text = md5
    return data_object


class Annotation:
    def __init__(self, burst_infos: Iterable[BurstInfo], metadata_type: str, image_number: int):
        """Initialize the Annotation object."""
        self.burst_infos = burst_infos
        self.metadata_type = metadata_type
        self.image_number = image_number
        self.metadata_paths = drop_duplicates([x.metadata_path for x in burst_infos])
        self.swath, self.pol = burst_infos[0].swath, burst_infos[0].polarization
        self.start_line = burst_infos[0].burst_index * burst_infos[0].length
        self.total_lines = len(burst_infos) * burst_infos[0].length
        self.stop_line = self.start_line + self.total_lines
        self.min_anx = min([x.start_utc for x in burst_infos])
        self.max_anx = max([x.stop_utc for x in burst_infos])

        self.inputs = [
            get_subxml_from_metadata(path, metadata_type, self.swath, self.pol) for path in self.metadata_paths
        ]

        products = [get_subxml_from_metadata(path, 'product', self.swath, self.pol) for path in self.metadata_paths]
        slc_lengths = []
        for annotation in products:
            n_bursts = int(annotation.find('.//burstList').get('count'))
            burst_length = int(annotation.find('.//linesPerBurst').text)
            slc_lengths.append(n_bursts * burst_length)
        self.slc_lengths = slc_lengths

        # annotation components to be extended by subclasses
        self.ads_header = None
        self.xml = None

        # these attributes are updated when the annotation is written to a file
        self.size_bytes = None
        self.md5 = None

    def create_ads_header(self):
        """Create the ADS header for the annotation."""
        ads_header = deepcopy(self.inputs[0].find('adsHeader'))
        ads_header.find('startTime').text = self.min_anx.isoformat()
        ads_header.find('stopTime').text = self.max_anx.isoformat()
        ads_header.find('imageNumber').text = f'{self.image_number:03d}'
        self.ads_header = ads_header

    def merge_lists(self, list_name: str, line_bounds: Optional[tuple[int, int]] = None) -> ET.Element:
        """Merge lists of elements into a single list.

        Args:
            list_name: The name of the list element.

        Returns:
            The merged list element.
        """
        list_elements = [input_xml.find(list_name) for input_xml in self.inputs]
        list_of_list_elements = ListOfListElements(list_elements, self.start_line, self.slc_lengths)
        merged_list = list_of_list_elements.create_filtered_list([self.min_anx, self.max_anx], line_bounds=line_bounds)
        return merged_list

    def write(self, out_path: Path, update_info=True) -> None:
        """Write the annotation to a file.

        Args:
            out_path: The path to write the annotation to.
            update_info: Whether to update the size and md5 attributes of the annotation.
        """
        self.xml.write(out_path, pretty_print=True, xml_declaration=True, encoding='utf-8')

        if update_info:
            self.path = out_path
            with open(out_path, 'rb') as f:
                file_bytes = f.read()
                self.size_bytes = len(file_bytes)
                self.md5 = hashlib.md5(file_bytes).hexdigest()

    def __str__(self, **kwargs):
        """Return the XML string representation of the annotation.

        Args:
            kwargs: Keyword arguments to pass to the lxml
        """
        xml_str = ET.tostring(self.xml, pretty_print=True, **kwargs)
        return xml_str.decode()

    def create_manifest_components(self):
        """Create the components for the manifest file."""
        unit_type = 'Metadata Unit'
        mime_type = 'text/xml'

        rep_id = f's1Level1{self.metadata_type.capitalize()}Schema'
        simple_name = self.path.with_suffix('').name.replace('-', '')
        if self.metadata_type == 'product':
            simple_name = f'product{simple_name}'

        safe_index = [i for i, x in enumerate(self.path.parts) if 'SAFE' in x][-1]
        safe_path = Path(*self.path.parts[: safe_index + 1])
        rel_path = self.path.relative_to(safe_path)

        content_unit = create_content_unit(simple_name, unit_type, rep_id)
        metadata_object = create_metadata_object(simple_name)
        data_object = create_data_object(simple_name, rel_path, rep_id, mime_type, self.size_bytes, self.md5)

        return content_unit, metadata_object, data_object



import lxml.etree as ET
import numpy as np
from osgeo import gdal, osr


gdal.UseExceptions()


def create_test_geotiff(output_file, dtype='float', value=1, shape=(10, 10, 1)):
    """Create a test geotiff for testing"""
    opts = {'float': (np.float64, gdal.GDT_Float64), 'cfloat': (np.complex64, gdal.GDT_CFloat32)}
    np_dtype, gdal_dtype = opts[dtype]
    data = np.ones(shape[0:2], dtype=np_dtype) * value
    geotransform = [0.0, 1.0, 0.0, 0.0, 0.0, 1.0]
    driver = gdal.GetDriverByName('GTiff')
    dataset = driver.Create(output_file, shape[1], shape[0], shape[2], gdal_dtype)
    dataset.SetGeoTransform(geotransform)
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(4326)
    dataset.SetProjection(srs.ExportToWkt())
    for i in range(shape[2]):
        band = dataset.GetRasterBand(i + 1)
        band.WriteArray(data)
    dataset = None


def validate_xml(xml_file, xsd_file):
    xml_doc = ET.parse(xml_file)

    xsd_doc = ET.parse(xsd_file)
    schema = ET.XMLSchema(xsd_doc)

    is_valid = schema.validate(xml_doc)
    if not is_valid:
        raise ValueError(f'XML failed validation with message:\n{schema.error_log}')

    return is_valid
